/**
 * data_pronouncer.cpp
 * ===================
 * Implementation of basic pronouncer functionality based on DATA Interactive.
 *
 * By Hkan L. Younes (lorens@acm.org)
 */

#include <malloc.h>
#include <stdio.h>

#include <ActiveDATA.h>

#include "data_pronouncer.h"


#define MAX_PATHSTRLEN (2*MAX_PATHLEN + 2)


/**
 * A ticket.
 */
static IUnknown* ticket = NULL;

/**
 * Path to ticket file.
 */
static char* ticket_path = "d:\\lorens\\exjobb\\src\\data\\sample.tkt";

/**
 * The models.
 */
static ITreeObj* model = NULL;

/**
 * Root nodes.
 */
static INodeObj* root_node = NULL;


static BSTR toBSTR(char *s)
{
    int len = lstrlen(s) + 1, doubleLen = len*2;
    LPWSTR lpw = (LPWSTR)alloca(doubleLen);
    *lpw = 0;
    MultiByteToWideChar(CP_ACP, 0, s, len, lpw, doubleLen);
    return SysAllocString(lpw);
}


static BSTR path_bstr(const int* path, int path_len)
{
    char path_str[MAX_PATHSTRLEN];

    path_str[0] = '$';
    for (int i = 0; i < path_len; i++)
    {
        path_str[2*i + 1] = (char)(path[i] + '1');
        path_str[2*(i + 1)] = ',';
    }
    path_str[2 * path_len] = '\0';

    return toBSTR(path_str);
}


static BSTR double_bstr(double x)
{
    char double_str[81];

    sprintf(double_str, "%f", x);

    return toBSTR(double_str);
}


/**
 * Initialize pronouncer.
 *
 * Returns -1 on failure, and 0 on success.
 *
 * N.B. must be called before the pronouncer is used.
 */
int initializePronouncer()
{
    HRESULT hr;
    IClassFactory* cf;
    ITicketObj* ti;
    IUnknown* tree;
    BSTR b;
    long err;

    CoInitialize(NULL);

    hr = CoGetClassObject(CLSID_TicketObj, CLSCTX_ALL, NULL,
							IID_IClassFactory, (void**)&cf);
    hr = cf->CreateInstance(NULL, IID_IUnknown, (void**)&ticket);
    cf->Release();
    hr = ticket->QueryInterface(IID_ITicketObj, (void**)&ti);
    b = toBSTR(ticket_path);
    hr = ti->CreateTicketFromFile(b, &err);
    SysFreeString(b);
    ti->Release();

    hr = CoGetClassObject(CLSID_TreeObj, CLSCTX_ALL, NULL,
							IID_IClassFactory, (void**)&cf);
    hr = cf->CreateInstance(NULL, IID_IUnknown, (void**)&tree);
    cf->Release();
    hr = tree->QueryInterface(IID_ITreeObj, (void**)&model);
    tree->Release();

    return (err == active_noErr) ? 0 : -1;
}


/**
 * Finalizes pronouncer.
 */
void finalizePronouncer()
{
    if (root_node != NULL)
        root_node->Release();
    model->Release();
    ticket->Release();
    CoUninitialize();
}


/**
 * Loads a model to work with.
 *
 * Parameters:
 *  path  - path to file containing the model.
 *
 * Returns -1 on failure, and 0 on success.
 */
int loadModel(char* path)
{
    HRESULT hr;
    BSTR b;
    long ok;

    b = toBSTR(path);
    if (root_node != NULL)
        root_node->Release();
#if 0
    hr = model->OpenPackage(b, ticket, &ok);
#else
    hr = model->OpenPackage(b, NULL, &ok);
#endif
    hr = model->SelectRoot();
    hr = model->GetNodeObj((IDispatch**)&root_node);

    return (ok) ? 0 : -1;
}


/**
 * Sets probabilities in the model.
 *
 * Parameters:
 *  node_path - path to node.
 *  path_len  - length of node path.
 *  probs     - array of probabilities to assign to row.
 *  nprobs    - number of probabilities in probs.
 */
void set_probabilities(const int* node_path, int path_len,
                       const double* probs, int nprobs)
{
    HRESULT hr;
    BSTR b;
    INodeObj* node;
    INodeObj* next_node;

    if (path_len == 0)
        node = root_node;
    else
    {
        b = path_bstr(node_path, path_len);
        hr = root_node->GetRelativeNode(b, (IDispatch**)&node);
        SysFreeString(b);
    }
    b = toBSTR("$1");
    hr = node->GetRelativeNode(b, (IDispatch**)&next_node);
    SysFreeString(b);
    if (node != root_node)
        node->Release();
    node = next_node;
    b = double_bstr(*probs++);
    hr = node->put_prob(b);
    SysFreeString(b);
    while (--nprobs)
    {
        b = toBSTR("$>");
        hr = node->GetRelativeNode(b, (IDispatch**)&next_node);
        SysFreeString(b);
        node->Release();
        node = next_node;
        b = double_bstr(*probs++);
        hr = node->put_prob(b);
        SysFreeString(b);
    }
    node->Release();
}


/**
 * Sets value of variable in the model.
 *
 * Parameters:
 *  node_path - path to node.
 *  path_len  - length of node path.
 *  var_name  - name of variable.
 *  value     - new value for variable.
 */
void set_variable(const int* node_path, int path_len,
                     char* var_name, double value)
{
    HRESULT hr;
    BSTR b;
    INodeObj* node;

    if (path_len == 0)
        node = root_node;
    else
    {
        b = path_bstr(node_path, path_len);
        hr = root_node->GetRelativeNode(b, (IDispatch**)&node);
        SysFreeString(b);
    }
    b = toBSTR(var_name);
    hr = node->SetVariable(b, value);
    SysFreeString(b);
    if (node != root_node)
        node->Release();
}


/**
 * Sets values of variables in the model.
 *
 * Parameters:
 *  node_path - path to node.
 *  path_len  - length of node path.
 *  values    - new values for variables.
 *  nvaluea   - number of variables to set.
 */
void setModelValues(const int* node_path, int path_len,
                       const double* values, int nvalues)
{
    HRESULT hr;
    BSTR b;
    INodeObj* node;
    IVarInfo* var_info;
    int i;

    if (path_len == 0)
        node = root_node;
    else
    {
        b = path_bstr(node_path, path_len);
        hr = root_node->GetRelativeNode(b, (IDispatch**)&node);
        SysFreeString(b);
    }
    for (i = 0; i < nvalues; i++)
    {
        hr = node->GetVarInfo((short)(i + 1), (IDispatch**)&var_info);
        hr = var_info->get_name(&b);
        var_info->Release();
        hr = node->SetVariable(b, *values++);
    }
    SysFreeString(b);
    if (node != root_node)
        node->Release();
}


/**
 * Gets an advice.
 *
 * Parameters:
 *  decision_path - path to decision node, or NULL for root node.
 *  path_len      - length of node path.
 *
 * Returns the index of the action that maximizes the expected utility.
 */
int evaluateModel(const int *decision_path, int path_len)
{
    HRESULT hr;
    BSTR b;
    INodeObj* node;
    long err;
    short advice;

    if (path_len == 0)
        node = root_node;
    else
    {
        b = path_bstr(decision_path, path_len);
        hr = root_node->GetRelativeNode(b, (IDispatch**)&node);
        SysFreeString(b);
    }
    hr = node->ExpVal(&err);
    hr = node->GetOptimal(&advice);
    if (node != root_node)
        node->Release();
#ifdef DEBUG
    {
        IClassFactory* cf;
        IUnknown* opt;
        IGraphicOptions* iopt;
        long err;

        hr = CoGetClassObject(CLSID_GraphicOptions, CLSCTX_ALL, NULL,
	    						IID_IClassFactory, (void**)&cf);
        hr = cf->CreateInstance(NULL, IID_IUnknown, (void**)&opt);
        cf->Release();
        hr = opt->QueryInterface(IID_IGraphicOptions, (void**)&iopt);
        b = toBSTR("d:\\lorens\\exjobb\\src\\data\\debug.bmp");
        hr = iopt->put_fileName(b);
        SysFreeString(b);
        iopt->Release();
        hr = model->SaveGraphic(opt, &err);
        opt->Release();
    }
#endif

    return advice - 1;
}
