/**
 * dte_pronouncer.c
 * ================
 * Implementation of basic pronouncer functionality based on DTE.
 *
 * By Hkan L. Younes (lorens@acm.org)
 */

#include <stdlib.h>

#include <dte.h>

#include "dte_pronouncer.h"


/**
 * Maximum number of models kept in memory.
 */
#define MAX_MODELS 10


/**
 * The models.
 */
static DecisionTree models[MAX_MODELS];


/**
 * Initialize pronouncer.
 *
 * Returns -1 on failure, and 0 on success.
 *
 * N.B. must be called before the pronouncer is used.
 */
int initializePronouncer(void)
{
    int i;

    for (i = 0; i < MAX_MODELS; i++)
	models[i] = NULL;

    return 0;
}


/**
 * Finalizes pronouncer.
 *
 * N.B. should be called after the last call to the pronouncer.
 */
void finalizePronouncer(void)
{
    int i;

    for (i = 0; i < MAX_MODELS; i++)
	freeModel(i);
}


/**
 * Loads a model.
 *
 * Parameters:
 *  path - path to file containing the model.
 *
 * Returns -1 on failure, and an identifier for the model on success.
 */
int loadModel(const char *path)
{
    int i;

    for (i = 0; i < MAX_MODELS && models[i] != NULL; i++)
	;
    if (i == MAX_MODELS)
	return -1;   /* maximum number of models loaded */

    models[i] = readTree(path);

    return (models[i] == NULL) ? -1 : i;
}


/**
 * Frees a model.
 *
 * Parameters:
 *  modelId - a model identifier as returned by loadModel.
 */
void freeModel(int modelId)
{
    if (models[modelId] != NULL)
	freeTree(models[modelId]);
}


/**
 * Sets values in the given model.
 *
 * Parameters:
 *  modelId - a model identifier as returned by loadModel.
 *  first   - index of first value to set.
 *  nvalues - number of values to set.
 *  values  - the values.
 */
void setModelValues(int modelId, int first, int nvalues,
		    const double *values)
{
    setValues(models[modelId], first, nvalues, values);
}


/**
 * Evaluates the given model.
 *
 * Parameters:
 *  modelId - a model identifier as returned by loadModel.
 *
 * Returns the index of an alternative that maximizes the expected utility.
 */
int evaluateModel(int modelId)
{
    return evaluateTree(models[modelId]);
}
