/**
 * smile_pronouncer.cc
 * ===================
 * Implementation of basic pronouncer functionality based on SMILE.
 *
 * By Hkan L. Younes (lorens@acm.org)
 */

#include <stdlib.h>

#include <network.h>

#include "smile_pronouncer.h"


/**
 * Maximum number of models kept in memory.
 */
#define MAX_MODELS 10

/**
 * Maximum number of parents that a node can have.
 */
#define MAX_PARENTS 32


/**
 * The models.
 */
static DSL_network models[MAX_MODELS];


/**
 * Initialize pronouncer.
 *
 * Returns -1 on failure, and 0 on success.
 *
 * N.B. must be called before the pronouncer is used.
 */
int initializePronouncer()
{
    return 0;
}


/**
 * Finalize pronouncer.
 *
 * N.B. should be called after the last call to the pronouncer.
 */
void finalizePronouncer()
{
    /* nothing to do */
}


/**
 * Loads a model.
 *
 * Parameters:
 *  path  - path to file containing the model.
 *
 * Returns -1 on failure, and an identifier for the model on success.
 */
int loadModel(char *path)
{
    int i;
    for (i = 0; i < MAX_MODELS && models[i].GetNumberOfNodes() != 0; i++)
	;
    if (i == MAX_MODELS)
	return -1;   /* maximum number of models loaded */

    int status = models[i].ReadFile(path);
    if (status != DSL_OKAY)
	return -1;

    models[i].SetDefaultIDAlgorithm(DSL_ALG_ID_SHACHTER);

    return 0;
}


/**
 * Frees a model.
 *
 * Parameters:
 *  modelId - a model identifier as returned by loadModel.
 */
void freeModel(int modelId)
{
    DSL_network newModel;
    models[modelId] = newModel;
}


int saveModel(int modelId, char *path)
{
    models[modelId].WriteFile(path);
    return 0;
}


/**
 * Sets probabilities in the given model.
 *
 * Parameters:
 *  modelId      - a model identifier as returned by loadModel.
 *  nodeName     - name of node.
 *  parentStates - states of parents indicating row in probability table.
 *  nprobs       - number of probabilities in probs.
 *  probs        - array of probabilities to assign to row.
 */
void setProbabilities(int modelId, char *nodeName,
		      const int *parentStates, int nprobs,
		      const double *probs)
{
    int node = models[modelId].FindNode(nodeName);
    DSL_Dmatrix* table;
    models[modelId].GetNode(node)->Definition()->GetDefinition(&table);
    int n = models[modelId].NumParents(node);
    DSL_intArray coords(n + 1);
	int i;
    for (i = 0; i < n; i++)
	coords.Add(parentStates[i]);
    coords.Add(0);
    for (i = 0; i < nprobs; i++)
    {
	(*table)[coords] = probs[i];
	coords[n]++;
    }
}


/**
 * Sets a utility in the given model.
 *
 * Parameters:
 *  modelId      - a model identifier as returned by loadModel.
 *  nodeName     - name of node.
 *  parentStates - states of parents indicating position in utility table.
 *  util         - the utility.
 */
void setUtility(int modelId, char *nodeName,
		const int *parentStates, double util)
{
    int node = models[modelId].FindNode(nodeName);
    DSL_Dmatrix* table;
    models[modelId].GetNode(node)->Definition()->GetDefinition(&table);
    int n = models[modelId].NumParents(node);
    DSL_intArray coords(n);
    for (int i = 0; i < n; i++)
	coords.Add(parentStates[i]);
    coords.Add(0);
    (*table)[coords] = util;
}


/**
 * Sets all probabilities in a node of the given model.
 *
 * Parameters:
 *  modelId       - a model identifier as returned by loadModel.
 *  nodeName      - name of node.
 *  nparents      - number of parents
 *  parentNstates - number of states that each of the parents can be in.
 *  nstates       - number of states that this node can be in.
 *  probs         - probabilities.
 */
void setAllProbabilities(int modelId, char *nodeName,
			 int nparents, const int *parentNstates,
			 int nstates, const double *probs)
{
    int parentStates[MAX_PARENTS];
    int i;

    for (i = 0; i < nparents - 1; i++)
	parentStates[i] = 0;
    parentStates[i] = -1;
    while (i >= 0)
    {
	i = nparents - 1;
	while (++parentStates[i] == parentNstates[i] && i >= 0)
	{
	    parentStates[i] = 0;
	    i--;
	    if (i < 0)
		return;
	}
	if (i >= 0)
	{
	    setProbabilities(modelId, nodeName, parentStates, nstates, probs);
	    probs += nstates;
	}
    }
}


/**
 * Sets all utilities in a node of the given model.
 *
 * Parameters:
 *  modelId       - a model identifier as returned by loadModel.
 *  nodeName      - name of node.
 *  nparents      - number of parents
 *  parentNstates - number of states that each of the parents can be in.
 *  utils         - utilities.
 */
void setAllUtilities(int modelId, char *nodeName, 
		     int nparents, const int *parentNstates,
		     const double *utils)
{
    int parentStates[MAX_PARENTS];
    int i;

    for (i = 0; i < nparents - 1; i++)
	parentStates[i] = 0;
    parentStates[i] = -1;
    while (i >= 0)
    {
	i = nparents - 1;
	while (++parentStates[i] == parentNstates[i] && i >= 0)
	{
	    parentStates[i] = 0;
	    i--;
	    if (i < 0)
		return;
	}
	if (i >= 0)
	{
	    setUtility(modelId, nodeName, parentStates, *utils);
	    utils++;
	}
    }
}


/**
 * Evaluates the given model.
 *
 * Parameters:
 *  modelId      - a model identifier as returned by loadModel.
 *  decisionName - name of decision node.
 *
 * Returns the index of an alternative that maximizes the expected utility.
 */
int evaluateModel(int modelId, char *decisionName)
{
    models[modelId].UpdateBeliefs();
    int handle = models[modelId].FindNode(decisionName);
    DSL_node* node = models[modelId].GetNode(handle);
    DSL_listOfDecisions* utils = (DSL_listOfDecisions*)node->Value();
    DSL_Dmatrix* policyValues;
    utils->GetValue(&policyValues);
    for (int i = 0; i < policyValues->GetSize(); i++)
    {
	if ((*policyValues)[i] == 1.0)
	    return i;
    }
    return 0;
}
